#define _BSD_SOURCE

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <netdb.h>
#include <mosquitto.h>
#include <time.h>
#include <string.h>
#include <curl/curl.h>
#include <errno.h>
#include <libpq-fe.h>

const char *MOSQUITTO_HOST = ""; // TODO VYPLNIT
const int MOSQUITTO_PORT = -1; // TODO VYPLNIT
const char *MOSQUITTO_USERNAME = ""; // TODO VYPLNIT
const char *MOSQUITTO_PASSWORD = ""; // TODO VYPLNIT
const int MOSQUITTO_KEEPALIVE = 900;
const bool MOSQUITTO_CLEAN_SESSION = true;

char CURL_SERVER_URL[64];

const int CHARS_PER_LINE = 72;
#define SIZE 100

const int DEVICE_TYPE_AT_V1 = 0x01;
const int DEVICE_TYPE_METEO_V1 = 0x02;
const int DEVICE_TYPE_CO2_V1 = 0x03;

const int NOISE_AVG_ID_HEX = 0x11; // SNOG sensor ID (hex)
const int NOISE_MAX_ID_HEX = 0x12; // SNOG sensor ID (hex)
const int DUST1_ID_HEX = 0x13; // SNOG sensor ID (hex)
const int DUST2_5_ID_HEX = 0x14; // SNOG sensor ID (hex)
const int DUST10_ID_HEX = 0x15; // SNOG sensor ID (hex)
const int VOC_ID_HEX = 0x05; // SNOG sensor ID (hex)
const int WIND_SPEED_ID_HEX = 0x09; // SNOG sensor ID (hex)
const int WIND_ANGLE_ID_HEX = 0x0a; // SNOG sensor ID (hex)
const int PRESS_ID_HEX = 0x0b; // SNOG sensor ID (hex)
const int TEMP_ID_HEX = 0x07; // SNOG sensor ID (hex)
const int HUMID_ID_HEX = 0x08; // SNOT sensor ID (hex)
const int CO2_AVG_ID_HEX = 0x17; // SNOG sensor ID (hex)
const int CO2_MAX_ID_HEX = 0x18; // SNOG sensor ID (hex)
const int CO2_LAST_ID_HEX = 0x16; // SNOG sensor ID (hex)

uint16_t ths_int=500;
uint8_t delay=5;
uint8_t int_cfg=0x25;


/** PG SQL SECTION **/
PGconn *conn;
char hostbuffer[256];


/** LORA MESSAGES FORMAT **/
typedef struct PacketHeader{
	uint8_t fw;
	uint8_t deviceType;
}PacketHeader;

typedef struct StatusPacket {
	PacketHeader header;
	uint16_t status;
} StatusPacket;

/* 
 * Message structure for AirTracker
 */
typedef struct MessageAT{
	PacketHeader header;
	uint8_t pm1;
	uint8_t pm2_5;
	uint8_t pm10;
	uint16_t noiseAvg;
	uint16_t noiseMax;
	uint16_t voc;
}__attribute__((__packed__))MessageAT;

/* 
 * Message structure for Meteo
 */
typedef struct MessageMeteo {
	PacketHeader header;
	uint16_t windSpeed;
    uint16_t windDirection;
    uint16_t press;
    int16_t temperature;
    uint16_t humidity;
}__attribute__((__packed__))MessageMeteo;

/* 
 * Message structure for CO2
 */
typedef struct MessageCO2 {
	PacketHeader header;
	uint16_t co2Avg;
	uint16_t co2Max;
	uint16_t co2Last;
	uint16_t press;
	int16_t temperature;
	uint16_t humidity;
}__attribute__((__packed__))MessageCO2;

/** BASE64 SECTION **/
typedef enum {
	step_a, step_b, step_c, step_d
} base64_decodestep;

typedef struct {
	base64_decodestep step;
	char plainchar;
} base64_decodestate;

typedef enum {
	step_A, step_B, step_C
} base64_encodestep;

typedef struct
{
	base64_encodestep step;
	char result;
	int stepcount;
} base64_encodestate;


int base64_decode_value(char value_in)
{
	static const char decoding[] = {62,-1,-1,-1,63,52,53,54,55,56,57,58,59,60,61,-1,-1,-1,-2,-1,-1,-1,0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,-1,-1,-1,-1,-1,-1,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50,51};
	static const char decoding_size = sizeof(decoding);
	value_in -= 43;
	//if (value_in < 0 || value_in >= decoding_size) return -1;
	if (value_in >= decoding_size) return -1;
	return decoding[(int)value_in];
}

void base64_init_decodestate(base64_decodestate* state_in)
{
	state_in->step = step_a;
	state_in->plainchar = 0;
}

int base64_decode_block(const char* code_in, const int length_in, char* plaintext_out, base64_decodestate* state_in)
{
	const char* codechar = code_in;
	char* plainchar = plaintext_out;
	char fragment;

	*plainchar = state_in->plainchar;

	switch (state_in->step)
	{
		while (1)
		{
			case step_a:
				do {
					if (codechar == code_in+length_in)
					{
						state_in->step = step_a;
						state_in->plainchar = *plainchar;
						return plainchar - plaintext_out;
					}
					fragment = (char)base64_decode_value(*codechar++);
				} while (fragment < 0);
				*plainchar    = (fragment & 0x03f) << 2;
			case step_b:
				do {
					if (codechar == code_in+length_in)
					{
						state_in->step = step_b;
						state_in->plainchar = *plainchar;
						return plainchar - plaintext_out;
					}
					fragment = (char)base64_decode_value(*codechar++);
				} while (fragment < 0);
				*plainchar++ |= (fragment & 0x030) >> 4;
				*plainchar    = (fragment & 0x00f) << 4;
			case step_c:
				do {
					if (codechar == code_in+length_in)
					{
						state_in->step = step_c;
						state_in->plainchar = *plainchar;
						return plainchar - plaintext_out;
					}
					fragment = (char)base64_decode_value(*codechar++);
				} while (fragment < 0);
				*plainchar++ |= (fragment & 0x03c) >> 2;
				*plainchar    = (fragment & 0x003) << 6;
			case step_d:
				do {
					if (codechar == code_in+length_in)
					{
						state_in->step = step_d;
						state_in->plainchar = *plainchar;
						return plainchar - plaintext_out;
					}
					fragment = (char)base64_decode_value(*codechar++);
				} while (fragment < 0);
				*plainchar++   |= (fragment & 0x03f);
		}
	}
	/* control should not reach here */
	return plainchar - plaintext_out;
}

void base64_init_encodestate(base64_encodestate* state_in)
{
	state_in->step = step_A;
	state_in->result = 0;
	state_in->stepcount = 0;
}

char base64_encode_value(char value_in)
{
	static const char* encoding = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
	if (value_in > 63) return '=';
	return encoding[(int)value_in];
}

int base64_encode_block(const char* plaintext_in, int length_in, char* code_out, base64_encodestate* state_in)
{
	const char* plainchar = plaintext_in;
	const char* const plaintextend = plaintext_in + length_in;
	char* codechar = code_out;
	char result;
	char fragment;

	result = state_in->result;

	switch (state_in->step)
	{
		while (1)
		{
			case step_A:
				if (plainchar == plaintextend)
				{
					state_in->result = result;
					state_in->step = step_A;
					return codechar - code_out;
				}
				fragment = *plainchar++;
				result = (fragment & 0x0fc) >> 2;
				*codechar++ = base64_encode_value(result);
				result = (fragment & 0x003) << 4;
			case step_B:
				if (plainchar == plaintextend)
				{
					state_in->result = result;
					state_in->step = step_B;
					return codechar - code_out;
				}
				fragment = *plainchar++;
				result |= (fragment & 0x0f0) >> 4;
				*codechar++ = base64_encode_value(result);
				result = (fragment & 0x00f) << 2;
			case step_C:
				if (plainchar == plaintextend)
				{
					state_in->result = result;
					state_in->step = step_C;
					return codechar - code_out;
				}
				fragment = *plainchar++;
				result |= (fragment & 0x0c0) >> 6;
				*codechar++ = base64_encode_value(result);
				result  = (fragment & 0x03f) >> 0;
				*codechar++ = base64_encode_value(result);

				++(state_in->stepcount);
				if (state_in->stepcount == CHARS_PER_LINE/4)
				{
					*codechar++ = '\n';
					state_in->stepcount = 0;
				}
		}
	}
	/* control should not reach here */
	return codechar - code_out;
}

int base64_encode_blockend(char* code_out, base64_encodestate* state_in)
{
	char* codechar = code_out;

	switch (state_in->step)
	{
		case step_B:
			*codechar++ = base64_encode_value(state_in->result);
			*codechar++ = '=';
			*codechar++ = '=';
			break;
		case step_C:
			*codechar++ = base64_encode_value(state_in->result);
			*codechar++ = '=';
			break;
		case step_A:
			break;
	}
	*codechar++ = '\n';

	return codechar - code_out;
}

char* encode(const char* input,char *output, int length)
{
	/* set up a destination buffer large enough to hold the encoded data */
	//char* output = (char*)malloc(SIZE);
	/* keep track of our encoded position */
	char* c = output;
	/* store the number of bytes encoded by a single call */
	int cnt = 0;
	/* we need an encoder state */
	base64_encodestate s;

	/*---------- START ENCODING ----------*/
	/* initialise the encoder state */
	base64_init_encodestate(&s);
	/* gather data from the input and send it to the output */
	cnt = base64_encode_block(input, length, c, &s);
	c += cnt;
	/* since we have encoded the entire input string, we know that
		 there is no more input data; finalise the encoding */
	cnt = base64_encode_blockend(c, &s);
	c += cnt;
	/*---------- STOP ENCODING  ----------*/

	/* we want to print the encoded data, so null-terminate it: */
	c--;
	*c = 0;
	//    printf("encoded string: %s, encoded cnt: %d\n", output, strlen(output));

	return output;
}

char* decode(const char* input,char *output, int length)
{
	/* set up a destination buffer large enough to hold the encoded data */
	//char* output = (char*)malloc(20);
	for (int i = 0; i < 20; i++) {
		output[i] = 0;
	}
	/* keep track of our decoded position */
	char* c = output;
	/* store the number of bytes decoded by a single call */
	int cnt = 0;
	/* we need a decoder state */
	base64_decodestate s;

	/*---------- START DECODING ----------*/
	/* initialise the decoder state */
	base64_init_decodestate(&s);
	/* decode the input data */
	cnt = base64_decode_block(input, length, c, &s);
	//    printf("decoding: %s, decoded cnt: %zu\n", input, cnt);
	c += cnt;
	/* note: there is no base64_decode_blockend! */
	/*---------- STOP DECODING  ----------*/

	/* we want to print the decoded data, so null-terminate it: */
	*c = 0;

	return output;
}

/** DATA CONVERSION **/
int hexchr2bin(const char hex, char *out) {
	if (out == NULL) return 0;
	if (hex >= '0' && hex <= '9') {
		*out = hex - '0';
	} else if (hex >= 'A' && hex <= 'F') {
		*out = hex - 'A' + 10;
	} else if (hex >= 'a' && hex <= 'f') {
		*out = hex - 'a' + 10;
	} else {
		return 0;
	}

	return 1;
}

size_t hexa2bytes(const char *val, unsigned char *out) {
	int length = strlen(val);
	for (int i = 0, index = 0; i < length; i+=2) {
		out[index] = (2<<4) * val[i] +  val[i+1];
	}
	return length/2;
}

size_t date2bytes(unsigned char *out) {
	time_t rawtime;
	struct tm *timeinfo;
	
	time (&rawtime);
	timeinfo = gmtime (&rawtime);

	out[0] = (int) timeinfo->tm_year-100;
	out[1] = timeinfo->tm_mon + 1;
	out[2] = timeinfo->tm_mday;
	out[3] = timeinfo->tm_hour;
	out[4] = timeinfo->tm_min;
	out[5] = timeinfo->tm_sec;
	
	return sizeof(out);
}

unsigned int hexa2Dec(char *hex) {
	int decimal = 0;
	int base = 1;
    int length = strlen(hex);
    for(int i = length--; i >= 0; i--)
    {
        if(hex[i] >= '0' && hex[i] <= '9')
        {
            decimal += (hex[i] - 48) * base;
            base *= 16;
        }
        else if(hex[i] >= 'A' && hex[i] <= 'F')
        {
            decimal += (hex[i] - 55) * base;
            base *= 16;
        }
        else if(hex[i] >= 'a' && hex[i] <= 'f')
        {
            decimal += (hex[i] - 87) * base;
            base *= 16;
        }
    }
	return decimal;
}

void encodeValueFromFloat(float f, unsigned char *out) {
	unsigned char bytes_array[4];
	union {
		float f;
		unsigned char tmp_array[4];
	} f2u = { .f = f };
	memcpy(bytes_array, f2u.tmp_array, 4);
	for (int i = 0; i < 4; i++) {
		out[3-i] = bytes_array[i];
	}
}

void encodeValueFromInt(uint16_t val, unsigned char *out) {
	unsigned char bytes_array[4];
	float f = (float) val;
	union {
		float f;
		unsigned char tmp_array[4];
	} f2u = { .f = f };
	memcpy(bytes_array, f2u.tmp_array, 4);
	for (int i = 0; i < 4; i++) {
		out[3-i] = bytes_array[i];
	}
}

void encodeValueFromHex(char *hex, unsigned char *out) {
	unsigned char bytes_array[4];
	float f = (float) hexa2Dec(hex);
	union {
		float f;
		unsigned char tmp_array[4];
	} f2u = { .f = f };
	memcpy(bytes_array, f2u.tmp_array, 4);
	for (int i = 0; i < 4; i++) {
		out[3-i] = bytes_array[i];
	}
}

void decimalTo2bytes(int number, unsigned char *out) {
	out[0] = (number >> 8);
	out[1] = (number & 0xff);
}

uint16_t swap2Bytes(uint16_t num) {
	return (num >> 8) | (num << 8);
}

uint32_t swap4Bytes(uint32_t num) {
	return ((num>>24)&0xff) | // move byte 3 to byte 0
		((num<<8)&0xff0000) | // move byte 1 to byte 2
		((num>>8)&0xff00) | // move byte 2 to byte 1
		((num<<24)&0xff000000); // byte 0 to byte 3
}

void checkHostName(int hostname) {
	if (hostname == -1) {
		perror("gethostname");
		exit(1);
	}
}

void printPacket(unsigned char *packet, int length) {
	printf("Packet: ");
	for (int i = 0; i < length; i++) {
		printf("%02x", packet[i]);
	} 
	printf("\n");
}

void printStatusPacket(StatusPacket *statusPacket) {
	printf("Status packet: %#04x\n", statusPacket->status);
}

void printMessageAT(int sensorId, int pointId, MessageAT *messageAT) {
	unsigned int i;
	(void) i;
	printf("Message AT (sensorId: %d, pointId: %d): payload: ", sensorId, pointId);
	for(i=0;i<sizeof(MessageAT);i++){
		printf("%#02x ",*(((unsigned char*)messageAT)+i));
	}
	printf("\nMessageAT: pm1: %u; pm2_5: %u, pm10: %u, 10 * noiseAvg: %u, 10 * noiseMax: %u, voc: %u\n",
			messageAT->pm1, messageAT->pm2_5, messageAT->pm10, messageAT->noiseAvg, messageAT->noiseMax, messageAT->voc);
}

void printMessageMeteo(int sensorId, int pointId, MessageMeteo *messageMeteo) {
	unsigned int i;
	(void) i;
	printf("Message Meteo: (sensorId: %d, pointId: %d): payload: ", sensorId, pointId);
	for(i=0;i<sizeof(MessageMeteo);i++){
		printf("%#02x ",*(((unsigned char*)messageMeteo)+i));
	}
	printf("\nMessageMeteo: windSpeed: %f; windDir: %u, press: %f, temp: %f, humidity: %f\n",
			messageMeteo->windSpeed/100.0, messageMeteo->windDirection, messageMeteo->press/10.0, messageMeteo->temperature/100.0, messageMeteo->humidity/100.0);
}

void printMessageCO2(int sensorId, int pointId, MessageCO2 *messageCO2) {
	unsigned int i;
	(void) i;
	printf("Message CO2(sensorId: %d, pointId: %d): payload: ", sensorId, pointId);
	for(i=0;i<sizeof(MessageCO2);i++){
		printf("%#02x ",*(((unsigned char*)messageCO2)+i));
	}
	printf("\nMessageCO2: co2Avg: %u; co2Max: %u, co2Last: %u, press: %f, temp: %f, humidity: %f\n",
			messageCO2->co2Avg, messageCO2->co2Max, messageCO2->co2Last,messageCO2->press/10.0, messageCO2->temperature/100.0, messageCO2->humidity/100.0);
}

void storeDataATToDB(char *devuid, int sensorId, int pointId, int rssi, MessageAT *messageAT) {
	if (conn == NULL) return;
	char query[512] = { };
	sprintf(query, "INSERT into master_log(sigfox_id, point_id, sensor_id, date, noise_avg, noise_max, dust_1, dust_2_5, dust_10, voc, rssi, lqi) values ('%s', '%d', '%d', now(), %f, %f, %d, %d, %d, %f, %d, '%s')",
		devuid, pointId, sensorId, messageAT->noiseAvg / 10.0, messageAT->noiseMax / 10.0, messageAT->pm1, messageAT->pm2_5, messageAT->pm10, (float) messageAT->voc, rssi, "LORA");
	printf("PG SQL query: %s\n", query);
	PGresult *res = PQexec(conn, query);
	if (PQresultStatus(res) != PGRES_COMMAND_OK) {
		printf("Error while inserting data to DB!\n");
	} else {
		printf("Data inserted to DB successfully.\n");
	}
	PQclear(res);
}

void storeDataMeteoToDB(char *devuid, int sensorId, int pointId, int rssi, MessageMeteo *messageMeteo) {
	if (conn == NULL) return;
	char query[512] = { };
	sprintf(query, "INSERT into meteo_log(sigfox_id, point_id, sensor_id, date, wind_speed, wind_direction, press, temperature, humidity, rssi, lqi) values ('%s', '%d', '%d', now(), %f, %f, %f, %f, %f, %d, '%s')",
		devuid, pointId, sensorId, messageMeteo->windSpeed/100.0, (float) messageMeteo->windDirection, messageMeteo->press/10.0, messageMeteo->temperature/100.0, messageMeteo->humidity/100.0, rssi, "LORA");
	printf("PG SQL query: %s\n", query);
	PGresult *res = PQexec(conn, query);
	if (PQresultStatus(res) != PGRES_COMMAND_OK) {
		printf("Error while inserting data to DB!\n");
	} else {
		printf("Data inserted to DB successfully.\n");
	}
	PQclear(res);
}

void storeDataCO2ToDB(char *devuid, int sensorId, int pointId, int rssi, MessageCO2 *messageCO2) {
	if (conn == NULL) return;
	char query[512] = { };
	sprintf(query, "INSERT into co2_log(sigfox_id, point_id, sensor_id, date, co2Avg, co2Max, co2Last, press, temperature, humidity, rssi, lqi) values ('%s', '%d', '%d', now(), %d, %d, %d, %f, %f, %f, %d, '%s')",
		devuid, pointId, sensorId, messageCO2->co2Avg, messageCO2->co2Max, messageCO2->co2Last, messageCO2->press/10.0, messageCO2->temperature/100.0, messageCO2->humidity/100.0, rssi, "LORA");
	printf("PG SQL query: %s\n", query);
	PGresult *res = PQexec(conn, query);
	if (PQresultStatus(res) != PGRES_COMMAND_OK) {
		printf("Error while inserting data to DB!\n");
	} else {
		printf("Data inserted to DB successfully.\n");
	}
	PQclear(res);
}

void storeStatusPacketToDB(char *devuid, int sensorId, int pointId, int rssi, StatusPacket *statusPacket) {
	if (conn == NULL) return;
	char query[512] = { };
	sprintf(query, "INSERT into status_log(dev_id, point_id, sensor_id, date,type, rssi, data) values ('%s', '%d', '%d', now(), 'LORA', %d, '%#04x')",
		devuid, pointId, sensorId, rssi, statusPacket->status);
	printf("PG SQL query: %s\n", query);
	PGresult *res = PQexec(conn, query);
	if (PQresultStatus(res) != PGRES_COMMAND_OK) {
		printf("Error while inserting data to DB!\n");
	} else {
		printf("Data inserted to DB successfully.\n");
	}
	PQclear(res);

}

void sendPacketCURL(unsigned char *packet, int size) {
	CURL *curl;
	CURLcode res;
	curl_global_init(CURL_GLOBAL_ALL);
	curl = curl_easy_init();
	if (curl) {
		curl_easy_setopt(curl, CURLOPT_URL, CURL_SERVER_URL);
		curl_easy_setopt(curl, CURLOPT_USERAGENT, "Mozilla");
		curl_easy_setopt(curl, CURLOPT_POST, 1);
		curl_easy_setopt(curl, CURLOPT_POSTFIELDS, packet);
		curl_easy_setopt(curl, CURLOPT_POSTFIELDSIZE, size);
		curl_easy_setopt(curl, CURLOPT_SSL_VERIFYPEER, false);
		curl_easy_setopt(curl, CURLOPT_SSL_VERIFYHOST, false);
		curl_easy_setopt(curl, CURLOPT_FOLLOWLOCATION, true);
		printf("curl_easy_perform result: ");
		res = curl_easy_perform(curl);
		printf("\n");
		if (res != CURLE_OK) {
			printf("curl_easy_perform() failed: %s\n", curl_easy_strerror(res));
		} else {
			printf("curl_easy_perform() - data sent to %s.\n", CURL_SERVER_URL);
		}
		curl_easy_cleanup(curl);
	}
	curl_global_cleanup();
}

void setPacketHeader(unsigned char *packet, int pointId) {
	packet[0] = 0x00;
	packet[1] = 0x00;
	packet[2] = 0x44;
	packet[3] = 0x47;
	decimalTo2bytes(pointId, &packet[4]);
	packet[6] = 0x00;
	packet[7] = 0x64;
	packet[8] = 0x70;
	packet[9] = 0x49;
	date2bytes(&packet[10]);	
}

void uploadDataMeteoToSNOG(int pointId, MessageMeteo *messageMeteo) {
	int sensorsCnt = 5;
	int size = 17 + sensorsCnt * 6;
	unsigned char *packet = (unsigned char*) malloc((size) * sizeof(unsigned char)); // delka paketu je max. 80B

	setPacketHeader(packet, pointId);
    packet[16] = sensorsCnt;

	// wind speed
	decimalTo2bytes(WIND_SPEED_ID_HEX, &packet[17]);
	encodeValueFromFloat(messageMeteo->windSpeed/100.0, &packet[19]);
	// sind direction
	decimalTo2bytes(WIND_ANGLE_ID_HEX,&packet[23]);
	encodeValueFromInt(messageMeteo->windDirection, &packet[25]);
	// press
	decimalTo2bytes(PRESS_ID_HEX, &packet[29]);
	encodeValueFromFloat(messageMeteo->press/10.0, &packet[31]);
	// temperature
	decimalTo2bytes(TEMP_ID_HEX, &packet[35]);
	encodeValueFromFloat(messageMeteo->temperature/100.0, &packet[37]);
	// humidity
	decimalTo2bytes(HUMID_ID_HEX, &packet[41]);
	encodeValueFromFloat(messageMeteo->humidity/100.0, &packet[43]);
	
	packet[size] = 0x00;

	printPacket(packet, size);
	sendPacketCURL(packet, size);

	free(packet);
}

void uploadDataCO2ToSNOG(int pointId, MessageCO2 *messageCO2) {
	int sensorsCnt = 5;
	int size = 17 + sensorsCnt * 6;
	unsigned char *packet = (unsigned char*) malloc((size) * sizeof(unsigned char)); // delka paketu je max. 80B

	setPacketHeader(packet, pointId);
    packet[16] = sensorsCnt;

	// CO2 Avg
	decimalTo2bytes(CO2_AVG_ID_HEX, &packet[17]);
	encodeValueFromInt(messageCO2->co2Avg, &packet[19]);
	// CO2 Max
	decimalTo2bytes(CO2_MAX_ID_HEX,&packet[23]);
	encodeValueFromInt(messageCO2->co2Max, &packet[25]);
	// CO2 Last
	decimalTo2bytes(CO2_LAST_ID_HEX,&packet[29]);
	encodeValueFromInt(messageCO2->co2Last, &packet[33]);
	// press
	decimalTo2bytes(PRESS_ID_HEX, &packet[35]);
	encodeValueFromFloat(messageCO2->press/10.0, &packet[37]);
	// temperature
	decimalTo2bytes(TEMP_ID_HEX, &packet[41]);
	encodeValueFromFloat(messageCO2->temperature/100.0, &packet[43]);
	// humidity
	decimalTo2bytes(HUMID_ID_HEX, &packet[47]);
	encodeValueFromFloat(messageCO2->humidity/100.0, &packet[49]);
	
	packet[size] = 0x00;

	printPacket(packet, size);
	sendPacketCURL(packet, size);

	free(packet);
}

void uploadDataATToSNOG(int pointId, MessageAT *messageAT) {
	int sensorsCnt = 6;
	int size = 17 + sensorsCnt * 6;
 	unsigned char *packet = (unsigned char*) malloc((size) * sizeof(unsigned char)); // delka paketu je max. 80B
	 
	setPacketHeader(packet, pointId);	
	packet[16] = sensorsCnt;
	// noise AVG
	decimalTo2bytes(NOISE_AVG_ID_HEX, &packet[17]);
	encodeValueFromFloat(messageAT->noiseAvg / 10.0, &packet[19]);
	// noise MAX
	decimalTo2bytes(NOISE_MAX_ID_HEX, &packet[23]);
	encodeValueFromFloat(messageAT->noiseMax / 10.0, &packet[25]);
	// dust1
	decimalTo2bytes(DUST1_ID_HEX, &packet[29]);
	encodeValueFromInt(messageAT->pm1, &packet[31]);
	// dust2_5
	decimalTo2bytes(DUST2_5_ID_HEX, &packet[35]);
	encodeValueFromInt(messageAT->pm2_5, &packet[37]);
	// dust10
	decimalTo2bytes(DUST10_ID_HEX, &packet[41]);
	encodeValueFromInt(messageAT->pm10, &packet[43]);
	// voc
	decimalTo2bytes(VOC_ID_HEX, &packet[47]);
	encodeValueFromInt(messageAT->voc, &packet[49]);


	packet[size] = 0x00;
 		
	printPacket(packet, size);
	sendPacketCURL(packet, size);

	free(packet);
}

int getPointIdFromSensorId(int sensorId) {
	if (sensorId == 100) {
		return 711;
	} else if (sensorId == 5) {
		return 701;
	} else if (sensorId == 6) {
		return 702;
	} else if (sensorId == 7) {
		return 703;
	} else if (sensorId == 8) {
		return 704;
	} else if (sensorId == 9) {
		return 705;
	} else if (sensorId == 10) {
		return 706;
	} else if (sensorId == 11) {
		return 707;
	} else if (sensorId == 12) {
		return 708;
	} else if (sensorId == 15) {
		return 709;
	} else if (sensorId == 14) {
		return 710;
	} else if (sensorId == 1) {
		return 712;
	} else if (sensorId == 2) {
		return 713;
	}
	return 0;
}

int getSensorIdFromEUI(char *devEUI) {
	if (!strcmp(devEUI, "26f26f01100002bd")) return 5;
	if (!strcmp(devEUI, "26f26f01100002be")) return 6;
	if (!strcmp(devEUI, "26f26f01100002bf")) return 7;
	if (!strcmp(devEUI, "26f26f01100002c0")) return 8;
	if (!strcmp(devEUI, "26f26f01100002c1")) return 9;
	if (!strcmp(devEUI, "26f26f01100002c2")) return 10;
	if (!strcmp(devEUI, "26f26f01100002c3")) return 11;
	if (!strcmp(devEUI, "26f26f01100002c4")) return 12;
	if (!strcmp(devEUI, "26f26f01100002c5")) return 15;
	if (!strcmp(devEUI, "26f26f01100002c7")) return 100;
	if (!strcmp(devEUI, "26f26f02100002c6")) return 14;
	if (!strcmp(devEUI, "26f26f03100002c8")) return 1;
	if (!strcmp(devEUI, "26f26f03100002c9")) return 2;
	return 0;
}

void mosq_message_callback(struct mosquitto *mosq, void *userdata, const struct mosquitto_message *message) {
	int res=0;
	(void) userdata;
	(void)res;
	printf("\nGot message %s for topic '%s'\nMid: %d, qos: %d, retain: %d\n", (char*) message->payload, message->topic, message->mid, message->qos, message->retain);

	char *msg = (char *) message->payload;
	char *found = strstr((char *) message->payload, "\"dataFrame\"");
	int payload_size;
	if (found != NULL) {
		found = found + 13;
		char *ind = strstr(found, "\"");
		msg = (char *) malloc(ind-found+1);
		strncpy(msg, found, ind-found);
		msg[ind-found] = '\0';
		payload_size=ind-found;
	}
	printf("Data frame: %s\n", msg);

	

	char *fPort = "1";
	found = strstr((char *) message->payload, "\"port\"");
	if (found != NULL) {
		found = found + 7;
		char *ind = strstr(found, ",");
		fPort = (char *) malloc(ind - found + 1);
		strncpy(fPort, found, ind-found);
		fPort[ind-found] = '\0';
	}
	printf("Port: %s\n", fPort);

	int intRssi = 0;
	found = strstr((char *) message->payload, "\"rssi\"");
	if (found != NULL) {
		found += 7;
		char *ind = strstr(found, ",");
		char *rssi = (char *) malloc(ind - found + 1);
		strncpy(rssi, found, ind-found);
		rssi[ind-found]='\0';
		printf("RSSI: rssi: %s\n", rssi);
		sscanf(rssi, "%d", &intRssi);
		
		free(rssi);
	}

	char *decodedPayload =(char*)malloc(256);
	decode(msg,decodedPayload ,payload_size);
	
///	printf("decodedPayload: %s\n", decodedPayload);

	char *devEUI = (char *) malloc (32);
	found = strstr((char *) message->payload, "deveui");
	if (found != NULL) {
		found += 9;
		char *ind = strstr(found, "\"");
		devEUI = (char *) malloc(ind - found +1);
		strncpy(devEUI, found, ind-found);
		devEUI[ind-found] = '\0';
	}
	printf("DevEUI: %s\n", devEUI);

	int sensorId = getSensorIdFromEUI(devEUI);
	int pointId = getPointIdFromSensorId(sensorId);

	int port;
	sscanf(fPort, "%d", &port);

	if (port == 2) { // datova zprava
		PacketHeader *packetHeader = (PacketHeader *) decodedPayload;
		if (packetHeader->deviceType == DEVICE_TYPE_AT_V1) {
			MessageAT *messageAT = (MessageAT *) decodedPayload;
			printMessageAT(sensorId, pointId, messageAT);
			if (pointId > 0) {
				uploadDataATToSNOG(pointId, messageAT);
			}
			storeDataATToDB(devEUI, sensorId, pointId, intRssi, messageAT);
		} else if (packetHeader->deviceType == DEVICE_TYPE_METEO_V1) {
			MessageMeteo *messageMeteo = (MessageMeteo *) decodedPayload;
			printMessageMeteo(sensorId, pointId, messageMeteo);
			if (pointId > 0) {
				uploadDataMeteoToSNOG(pointId, messageMeteo);
			}	
			storeDataMeteoToDB(devEUI, sensorId, pointId, intRssi, messageMeteo);
		} else if (packetHeader->deviceType == DEVICE_TYPE_CO2_V1) {
			MessageCO2 *messageCO2 = (MessageCO2 *) decodedPayload;
			printMessageCO2(sensorId, pointId, messageCO2);
			if (pointId > 0) {
				uploadDataCO2ToSNOG(pointId, messageCO2);
			}
			storeDataCO2ToDB(devEUI, sensorId, pointId, intRssi, messageCO2);
		}
	} else if (port == 3) { // stavovy paket
		printf("!!! Stavovy paket !!!\n");
		StatusPacket *statusPacket = (StatusPacket *) decodedPayload;
		printStatusPacket(statusPacket);
		storeStatusPacketToDB(devEUI, sensorId, pointId, intRssi, statusPacket);
	}

	free(msg);
	free(fPort);
	free(decodedPayload);
	free(devEUI);
}

void mosq_connect_callback(struct mosquitto *mosq, void *userdata, int result)
{
	(void) userdata;
    printf("mosq_connect_callback\n");
	if(!result){
		/* Subscribe to broker information topics on successful connect. */
		printf("mosquitto topic subscription to admin@cvut.cz/payloads_ul\n");
		mosquitto_subscribe(mosq, NULL, "admin@cvut.cz/payloads_ul", 0);
	} else {
		fprintf(stderr, "Connect failed\n");
	}
}

void mosq_subscribe_callback(struct mosquitto *mosq, void *userdata, int mid, int qos_count, const int *granted_qos)
{
	(void) userdata;
	(void) mosq;

	int i;

	printf("Subscribed (mid: %d): %d", mid, granted_qos[0]);
	for(i=1; i<qos_count; i++){
		printf(", %d", granted_qos[i]);
	}
	printf("\n");
}

void mosq_log_callback(struct mosquitto *mosq, void *userdata, int level, const char *str)
{
	(void) mosq;
	(void) userdata;
	(void) level;
	/* Pring all log messages regardless of level. */
	//	printf("%s\n", str);
}

int main(int argc, char *argv[])
{
	struct mosquitto *mosq = NULL;
	char hostbuffer[64];
	int hostname = gethostname(hostbuffer, sizeof(hostbuffer));
	checkHostName(hostname);
	sprintf(CURL_SERVER_URL, "https://%s/sn/gauge", hostbuffer);

	printf("Start aplikace - CURL: %s.\n", CURL_SERVER_URL);

	conn = PQconnectdb("host=localhost user=units password=units dbname=units");
	if (PQstatus(conn) == CONNECTION_BAD) {
		printf("Connection to PGSQL database failed: %s]n", PQerrorMessage(conn));
		conn = NULL;
	}
	mosquitto_lib_init();
	mosq = mosquitto_new(hostbuffer, MOSQUITTO_CLEAN_SESSION, NULL);
	if(!mosq){
		fprintf(stderr, "Error: Out of memory!\n");
		return 1;
	}

	mosquitto_username_pw_set(mosq, MOSQUITTO_USERNAME, MOSQUITTO_PASSWORD);
	mosquitto_tls_set(mosq, "DigiCertGlobalRootCA.crt.pem", NULL, NULL, NULL, NULL); // TODO vyplnit spravne jmeno certifikatu
	mosquitto_tls_opts_set(mosq, 0, NULL, NULL);
	mosquitto_tls_insecure_set(mosq, true);

	mosquitto_log_callback_set(mosq, mosq_log_callback);
	mosquitto_connect_callback_set(mosq, mosq_connect_callback);
	mosquitto_message_callback_set(mosq, mosq_message_callback);
	mosquitto_subscribe_callback_set(mosq, mosq_subscribe_callback);

	errno = 0;
	int connState = mosquitto_connect(mosq, MOSQUITTO_HOST, MOSQUITTO_PORT, MOSQUITTO_KEEPALIVE);
 	if (connState == MOSQ_ERR_SUCCESS) {
		printf("Mosquitto - connected successfully.\n");
	} else if (connState == MOSQ_ERR_INVAL) {
		printf("Mosquitto - invalid input parameters, exiting!\n");
	} else {
		printf("Mosquitto - error on connection!\n");
//		strerror_r();
	}

	mosquitto_loop_forever(mosq, -1, 1);

	mosquitto_destroy(mosq);
	mosquitto_lib_cleanup();

	if (conn != NULL) PQfinish(conn);

	return 0;
}
